import numpy as np


from cp_measure.core.measureobjectintensitydistribution import get_radial_distribution

size = 240
rng = np.random.default_rng(42)
# pixels = rng.integers(low=1, high=255, size=(size, size)) # Random
pixels = np.ones((size, size), dtype=np.integer)  # All ones
pixels[70:80, 70:80] = 10
# Create two similar-sized objects
masks = np.zeros_like(pixels)
masks[50:100, 50:100] = 1  # First square 50x50
# masks[80:120, 90:120] = 1  # Major asymmetries on bottom right edge
masks[150:200, 150:200] = 2  # Second square 50x50
# masks[175:180, 180:210] = 2  # Minor asymmetries on bottom right edge

# Get measurements
result = get_radial_distribution(masks, pixels)



"""
Wrapper to fetch measurement functions in bulk. We are keeping the ugly filenames to
match the original names. This may change in the future.
"""

from typing import Callable


def get_core_measurements() -> dict[str, Callable]:
    from cp_measure.core import (
        measuregranularity,
        measureobjectintensity,
        measureobjectintensitydistribution,
        measureobjectsizeshape,
        measuretexture,
    )

    return {
        "radial_distribution": measureobjectintensitydistribution.get_radial_distribution,
        "radial_zernikes": measureobjectintensitydistribution.get_radial_zernikes,
        "intensity": measureobjectintensity.get_intensity,
        "sizeshape": measureobjectsizeshape.get_sizeshape,
        "zernike": measureobjectsizeshape.get_zernike,
        "ferret": measureobjectsizeshape.get_ferret,
        "texture": measuretexture.get_texture,
        "granularity": measuregranularity.get_granularity,
    }


def get_correlation_measurements() -> dict[str, Callable]:
    from cp_measure.core import measurecolocalization

    return {
        "costes": measurecolocalization.get_correlation_costes,
        "pearson": measurecolocalization.get_correlation_pearson,
        "manders_fold": measurecolocalization.get_correlation_manders_fold,
        "rwc": measurecolocalization.get_correlation_rwc,
    }


def get_multimask_measurements() -> dict[str, Callable]:
    from cp_measure.multimask.measureobjectneighbors import measureobjectneighbors
    from cp_measure.multimask.measureobjectoverlap import measureobjectoverlap

    return {
        "overlap": measureobjectoverlap,
        "neighbors": measureobjectneighbors,
    }



"""
MeasureGranularity
==================
**MeasureGranularity** outputs spectra of size measurements of the
textures in the image.

Image granularity is a texture measurement that tries to fit a series of
structure elements of increasing size into the texture of the image and outputs a spectrum of measures
based on how well they fit.
Granularity is measured as described by Ilya Ravkin (references below).

Basically, MeasureGranularity:
1 - Downsamples the image (if you tell it to). This is set in
**Subsampling factor for granularity measurements** or **Subsampling factor for background reduction**.
2 - Background subtracts anything larger than the radius in pixels set in
**Radius of structuring element.**
3 - For as many times as you set in **Range of the granular spectrum**, it gets rid of bright areas
that are only 1 pixel across, reports how much signal was lost by doing that, then repeats.
i.e. The first time it removes one pixel from all bright areas in the image,
(effectively deleting those that are only 1 pixel in size) and then reports what % of the signal was lost.
It then takes the first-iteration image and repeats the removal and reporting (effectively reporting
the amount of signal that is two pixels in size). etc.

============ ============ ===============
Supports 2D? Supports 3D? Respects masks?
============ ============ ===============
YES          YES          YES
============ ============ ===============

Measurements made by this module
^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^

-  *Granularity:* The module returns one measurement for each instance
   of the granularity spectrum set in **Range of the granular spectrum**.

References
^^^^^^^^^^

-  Serra J. (1989) *Image Analysis and Mathematical Morphology*, Vol. 1.
   Academic Press, London
-  Maragos P. “Pattern spectrum and multiscale shape representation”,
   *IEEE Transactions on Pattern Analysis and Machine Intelligence*, 11,
   N 7, pp. 701-716, 1989
-  Vincent L. (2000) “Granulometries and Opening Trees”, *Fundamenta
   Informaticae*, 41, No. 1-2, pp. 57-90, IOS Press, 2000.
-  Vincent L. (1992) “Morphological Area Opening and Closing for
   Grayscale Images”, *Proc. NATO Shape in Picture Workshop*,
   Driebergen, The Netherlands, pp. 197-208.
-  Ravkin I, Temov V. (1988) “Bit representation techniques and image
   processing”, *Applied Informatics*, v.14, pp. 41-90, Finances and
   Statistics, Moskow, (in Russian)
"""

import numpy
import scipy.ndimage
import skimage.morphology
from cp_measure.utils import _ensure_np_array as fix


def get_granularity(
    mask: numpy.ndarray,
    pixels: numpy.ndarray,
    subsample_size: float = 0.25,
    image_sample_size: float = 0.25,
    element_size: int = 10,
    granular_spectrum_length: int = 16,
) -> dict[str, float]:
    """
    1. (Outcommented) Subsample image
    2.  Remove background pixels using a greyscale tophat filter
    3.  Calculate granular spectrum (size distribution) for all masks

    Parameters
    ----------
    subsample_size : float, optional
        Subsampling factor for granularity measurements.
        If the textures of interest are larger than a few pixels, we recommend
        you subsample the image with a factor <1 to speed up the processing.
        Downsampling the image will let you detect larger structures with a
        smaller sized structure element. A factor >1 might increase the accuracy
        but also require more processing time. Images are typically of higher
        resolution than is required for granularity measurements, so the default
        value is 0.25. For low-resolution images, increase the subsampling
        fraction; for high-resolution images, decrease the subsampling fraction.
        Subsampling by 1/4 reduces computation time by (1/4) :sup:`3` because the
        size of the image is (1/4) :sup:`2` of original and the range of granular
        spectrum can be 1/4 of original. Moreover, the results are sometimes
        actually a little better with subsampling, which is probably because
        with subsampling the individual granular spectrum components can be used
        as features, whereas without subsampling a feature should be a sum of
        several adjacent granular spectrum components. The recommendation on the
        numerical value cannot be determined in advance; an analysis as in this
        reference may be required before running the whole set. See this `pdf`_,
        slides 27-31, 49-50.

        .. _pdf:     http://www.ravkin.net/presentations/Statistical%20properties%20of%20algorithms%20for%20analysis%20of%20cell%20images.pdf"

    image_sample_size : float, optional
        Subsampling factor for background reduction.
        It is important to remove low frequency image background variations as
        they will affect the final granularity measurement. Any method can be
        used as a pre-processing step prior to this module; we have chosen to
        simply subtract a highly open image. To do it quickly, we subsample the
        image first. The subsampling factor for background reduction is usually
        [0.125 – 0.25]. This is highly empirical, but a small factor should be
        used if the structures of interest are large. The significance of
        background removal in the context of granulometry is that image volume
        at certain granular size is normalized by total image volume, which
        depends on how the background was removed.

    element_size : int, optional
        Radius of structuring element.
        This radius should correspond to the radius of the textures of interest
        *after* subsampling; i.e., if textures in the original image scale have
        a radius of 40 pixels, and a subsampling factor of 0.25 is used, the
        structuring element size should be 10 or slightly smaller, and the range
        of the spectrum defined below will cover more sizes.

    granular_spectrum_length : int, optional
        Range of the granular spectrum.
        You may need a trial run to see which granular
        spectrum range yields informative measurements. Start by using a wide
        spectrum and narrow it down to the informative range to save time.

    Returns
    -------
    Dictionary of 1-d arrays where each value contains the specific granularity feature for a given cells.

    Examples

    """
    #
    # Downsample the image and mask
    #
    new_shape = numpy.array(pixels.shape)
    # MODIFIED: Remove subsample
    # if subsample_size < 1:
    #     new_shape = new_shape * subsample_size
    #     if pixels.ndim == 2:
    #         i, j = (
    #             numpy.mgrid[0 : new_shape[0], 0 : new_shape[1]].astype(float)
    #             / subsample_size
    #         )
    #         pixels = scipy.ndimage.map_coordinates(pixels, (i, j), order=1)
    #         mask = scipy.ndimage.map_coordinates(mask.astype(float), (i, j)) > 0.9
    #     else:
    #         k, i, j = (
    #             numpy.mgrid[
    #                 0 : new_shape[0], 0 : new_shape[1], 0 : new_shape[2]
    #             ].astype(float)
    #             / subsample_size
    #         )
    #         pixels = scipy.ndimage.map_coordinates(pixels, (k, i, j), order=1)
    #         mask = scipy.ndimage.map_coordinates(mask.astype(float), (k, i, j)) > 0.9
    # else:
    pixels = pixels.copy()
    mask = mask.copy()
    #
    # Remove background pixels using a greyscale tophat filter
    #
    if image_sample_size < 1:
        back_shape = new_shape * image_sample_size
        if pixels.ndim == 2:
            i, j = (
                numpy.mgrid[0 : back_shape[0], 0 : back_shape[1]].astype(float)
                / image_sample_size
            )
            back_pixels = scipy.ndimage.map_coordinates(pixels, (i, j), order=1)
            back_mask = scipy.ndimage.map_coordinates(mask.astype(float), (i, j)) > 0.9
        else:
            k, i, j = (
                numpy.mgrid[
                    0 : new_shape[0], 0 : new_shape[1], 0 : new_shape[2]
                ].astype(float)
                / subsample_size
            )
            back_pixels = scipy.ndimage.map_coordinates(pixels, (k, i, j), order=1)
            back_mask = (
                scipy.ndimage.map_coordinates(mask.astype(float), (k, i, j)) > 0.9
            )
    else:
        back_pixels = pixels
        back_mask = mask
        back_shape = new_shape
    radius = element_size
    if pixels.ndim == 2:
        footprint = skimage.morphology.disk(radius, dtype=bool)
    else:
        footprint = skimage.morphology.ball(radius, dtype=bool)
    back_pixels_mask = numpy.zeros_like(back_pixels)
    back_pixels_mask[back_mask == 1] = back_pixels[back_mask == 1]
    back_pixels = skimage.morphology.erosion(back_pixels_mask, footprint=footprint)
    back_pixels_mask = numpy.zeros_like(back_pixels)
    back_pixels_mask[back_mask == 1] = back_pixels[back_mask == 1]
    back_pixels = skimage.morphology.dilation(back_pixels_mask, footprint=footprint)
    if image_sample_size < 1:
        if pixels.ndim == 2:
            i, j = numpy.mgrid[0 : new_shape[0], 0 : new_shape[1]].astype(float)
            #
            # Make sure the mapping only references the index range of
            # back_pixels.
            #
            i *= float(back_shape[0] - 1) / float(new_shape[0] - 1)
            j *= float(back_shape[1] - 1) / float(new_shape[1] - 1)
            back_pixels = scipy.ndimage.map_coordinates(back_pixels, (i, j), order=1)
        else:
            k, i, j = numpy.mgrid[
                0 : new_shape[0], 0 : new_shape[1], 0 : new_shape[2]
            ].astype(float)
            k *= float(back_shape[0] - 1) / float(new_shape[0] - 1)
            i *= float(back_shape[1] - 1) / float(new_shape[1] - 1)
            j *= float(back_shape[2] - 1) / float(new_shape[2] - 1)
            back_pixels = scipy.ndimage.map_coordinates(back_pixels, (k, i, j), order=1)
    pixels -= back_pixels
    pixels[pixels < 0] = 0

    # Transcribed from the Matlab module: granspectr function
    #
    # CALCULATES GRANULAR SPECTRUM, ALSO KNOWN AS SIZE DISTRIBUTION,
    # GRANULOMETRY, AND PATTERN SPECTRUM, SEE REF.:
    # J.Serra, Image Analysis and Mathematical Morphology, Vol. 1. Academic Press, London, 1989
    # Maragos,P. "Pattern spectrum and multiscale shape representation", IEEE Transactions on Pattern Analysis and Machine Intelligence, 11, N 7, pp. 701-716, 1989
    # L.Vincent "Granulometries and Opening Trees", Fundamenta Informaticae, 41, No. 1-2, pp. 57-90, IOS Press, 2000.
    # L.Vincent "Morphological Area Opening and Closing for Grayscale Images", Proc. NATO Shape in Picture Workshop, Driebergen, The Netherlands, pp. 197-208, 1992.
    # I.Ravkin, V.Temov "Bit representation techniques and image processing", Applied Informatics, v.14, pp. 41-90, Finances and Statistics, Moskow, 1988 (in Russian)
    # THIS IMPLEMENTATION INSTEAD OF OPENING USES EROSION FOLLOWED BY RECONSTRUCTION
    #
    ng = granular_spectrum_length
    # startmean = numpy.mean(pixels[mask])
    startmean = numpy.mean(pixels)
    ero = pixels.copy()
    # Mask the test image so that masked pixels will have no effect
    # during reconstruction
    #
    # ero[~mask] = 0
    startmean = max(startmean, numpy.finfo(float).eps)

    if pixels.ndim == 2:
        footprint = skimage.morphology.disk(1, dtype=bool)
    else:
        footprint = skimage.morphology.ball(1, dtype=bool)

    unique_labels = numpy.unique(mask)
    unique_labels = unique_labels[unique_labels > 0]

    # Info on objects
    range_ = numpy.arange(1, numpy.max(mask) + 1)
    labels = mask.copy()
    current_mean = fix(scipy.ndimage.mean(pixels, labels, range_))
    start_mean = numpy.maximum(current_mean, numpy.finfo(float).eps)

    results = {}
    for granularity_id in range(1, ng + 1):
        # NOTE: This seems to be an iterative process of sequential
        # erosions and reconstructions
        # ero_mask = numpy.zeros_like(ero)
        # ero_mask[mask == True] = ero[mask == True]
        ero_mask = ero.copy()
        # Shrink bright regions
        ero = skimage.morphology.erosion(ero_mask, footprint=footprint)
        # Use a mask (footprint) to make bright sections bigger
        rec = skimage.morphology.reconstruction(ero, pixels, footprint=footprint)
        # currentmean = numpy.mean(rec[mask])
        # gs is the image granularity
        # gs = (prevmean - currentmean) * 100 / startmean
        # image_granularity = gs

        # Restore the reconstructed image to the shape of the
        # original image so we can match against object labels
        #
        orig_shape = pixels.shape
        # TODO DRY This can be easily cleaned up
        if pixels.ndim == 2:
            i, j = numpy.mgrid[0 : orig_shape[0], 0 : orig_shape[1]].astype(float)
            #
            # Make sure the mapping only references the index range of
            # back_pixels.
            #
            i *= float(new_shape[0] - 1) / float(orig_shape[0] - 1)
            j *= float(new_shape[1] - 1) / float(orig_shape[1] - 1)
            rec = scipy.ndimage.map_coordinates(rec, (i, j), order=1)
        else:
            k, i, j = numpy.mgrid[
                0 : orig_shape[0], 0 : orig_shape[1], 0 : orig_shape[2]
            ].astype(float)
            k *= float(new_shape[0] - 1) / float(orig_shape[0] - 1)
            i *= float(new_shape[1] - 1) / float(orig_shape[1] - 1)
            j *= float(new_shape[2] - 1) / float(orig_shape[2] - 1)
            rec = scipy.ndimage.map_coordinates(rec, (k, i, j), order=1)

        # Calculate the means for the objects
        gss = numpy.zeros((0,))
        if unique_labels.any():
            #
            # Calculate the means for the objects
            #
            # MODIFIED: These metrics were defined inside ObjectRecord originally

            new_mean = fix(scipy.ndimage.mean(rec, labels, range_))
            gss = (current_mean - new_mean) * 100 / start_mean

        results[f"Granularity_{granularity_id}"] = gss

    return results



from typing import Optional

import centrosome.cpmorphology
import centrosome.zernike
import numpy
import scipy.ndimage
import skimage.measure
from cp_measure.utils import masks_to_ijv, _ensure_np_scalar

__doc__ = """\
MeasureObjectSizeShape
======================

**MeasureObjectSizeShape** measures several area and shape features
of identified objects.

============ ============ ===============
Supports 2D? Supports 3D? Respects masks?
============ ============ ===============
YES          YES          NO
============ ============ ===============

See also
^^^^^^^^

See also **MeasureImageAreaOccupied**.

Measurements made by this module
^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^
Some measurements are available for 3D and 2D objects, while some are 2D
only.

See the *Technical Notes* below for an explanation of a key step
underlying many of the following metrics: creating an
ellipse with the same second-moments as each object.

-  *Area:* *(2D only)* The number of pixels in the region.
-  *Volume:* *(3D only)* The number of voxels in the region.
-  *Perimeter:* *(2D only)* The total number of pixels around the boundary of each
   region in the image.
-  *SurfaceArea:* *(3D only)* The total number of voxels around the boundary of
   each region in the image.
-  *FormFactor:* *(2D only)* Calculated as 4\*π\*Area/Perimeter\ :sup:`2`. Equals 1
   for a perfectly circular object.
-  *Convex Area:* The area of a convex polygon containing the whole object.
   Best imagined as a rubber band stretched around the object. 
-  *Solidity:* The proportion of the pixels in the convex hull that are
   also in the object, i.e., *ObjectArea/ConvexHullArea*.
-  *Extent:* The proportion of the pixels (2D) or voxels (3D) in the bounding box
   that are also in the region. Computed as the area/volume of the object divided
   by the area/volume of the bounding box.
-  *EulerNumber:* The number of objects in the region minus the number
   of holes in those objects, assuming 8-connectivity.
-  *Center\_X, Center\_Y, Center\_Z:* The *x*-, *y*-, and (for 3D objects) *z-*
   coordinates of the point farthest away from any object edge (the *centroid*).
   Note that this is not the same as the *Location-X* and *-Y* measurements
   produced by the **Identify** or **Watershed**
   modules or the *Location-Z* measurement produced by the **Watershed** module.
-  *BoundingBoxMinimum/Maximum\_X/Y/Z:* The minimum/maximum *x*-, *y*-, and (for 3D objects)
   *z-* coordinates of the object.
-  *BoundingBoxArea:* *(2D only)* The area of a box containing the object.
-  *BoundingBoxVolume:* *(3D only)* The volume of a box containing the object.
-  *Eccentricity:* *(2D only)* The eccentricity of the ellipse that has the same
   second-moments as the region. The eccentricity is the ratio of the
   distance between the foci of the ellipse and its major axis length.
   The value is between 0 and 1. (0 and 1 are degenerate cases; an
   ellipse whose eccentricity is 0 is actually a circle, while an
   ellipse whose eccentricity is 1 is a line segment.)
-  *MajorAxisLength:* The length (in pixels) of the major axis of the
   ellipse that has the same normalized second central moments as the
   region.
-  *MinorAxisLength:* The length (in pixels) of the minor axis of the
   ellipse that has the same normalized second central moments as the
   region.
-  *EquivalentDiameter:* The diameter of a circle or sphere with the same area
   as the object.
-  *Orientation:* *(2D only)* The angle (in degrees ranging from -90 to 90 degrees)
   between the x-axis and the major axis of the ellipse that has the
   same second-moments as the region.
-  *Compactness:* *(2D only)* Calculated as Perimeter\ :sup:`2`/4\*π\*Area, related to 
   Form Factor. A filled circle will have a compactness of 1, with irregular objects or 
   objects with holes having a value greater than 1.
-  *MaximumRadius:* *(2D only)* The maximum distance of any pixel in the object to
   the closest pixel outside of the object. For skinny objects, this is
   1/2 of the maximum width of the object.
-  *MedianRadius:* *(2D only)* The median distance of any pixel in the object to the
   closest pixel outside of the object.
-  *MeanRadius:* *(2D only)* The mean distance of any pixel in the object to the
   closest pixel outside of the object.
-  *MinFeretDiameter, MaxFeretDiameter:* *(2D only)* The Feret diameter is the
   distance between two parallel lines tangent on either side of the
   object (imagine taking a caliper and measuring the object at various
   angles). The minimum and maximum Feret diameters are the smallest and
   largest possible diameters, rotating the calipers along all possible
   angles.
-  *Zernike shape features:* *(2D only)* These metrics of shape describe a binary object
   (or more precisely, a patch with background and an object in the
   center) in a basis of Zernike polynomials, using the coefficients as
   features (*Boland et al., 1998*). Currently, Zernike polynomials from
   order 0 to order 9 are calculated, giving in total 30 measurements.
   While there is no limit to the order which can be calculated (and
   indeed you could add more by adjusting the code), the higher order
   polynomials carry less information.
-  *Spatial Moment features:* *(2D only)* A series of weighted averages 
   representing the shape, size, rotation and location of the object.
-  *Central Moment features:* *(2D only)* Similar to spatial moments, but
   normalized to the object's centroid. These are therefore not influenced
   by an object's location within an image.
-  *Normalized Moment features:* *(2D only)* Similar to central moments,
   but further normalized to be scale invariant. These moments are therefore
   not impacted by an object's size (or location).
-  *Hu Moment features:* *(2D only)* Hu's set of image moment features. These
   are not altered by the object's location, size or rotation. This means that
   they primarily describe the shape of the object.
-  *Inertia Tensor features:* *(2D only)* A representation of rotational
   inertia of the object relative to it's center.
-  *Inertia Tensor Eigenvalues features:* *(2D only)* Values describing 
   the movement of the Inertia Tensor array.



Technical notes
^^^^^^^^^^^^^^^

A number of the object measurements are generated by creating an ellipse
with the same second-moments as the original object region. This is
essentially the best-fitting ellipse for a given object with the same
statistical properties. Furthermore, they are not affected by the
translation or uniform scaling of a region.

Following computer vision conventions, the origin of the X and Y axes is at the top
left of the image rather than the bottom left; the orientation of objects whose topmost point
is on their right (or are rotated counter-clockwise from the horizontal) will therefore
have a negative orientation, while objects whose topmost point is on their left
(or are rotated clockwise from the horizontal) will have a positive orientation.

The Zernike features are computed within the minimum enclosing circle of
the object, i.e., the circle of the smallest diameter that contains all
of the object’s pixels.

References
^^^^^^^^^^

-  Rocha L, Velho L, Carvalho PCP, “Image moments-based structuring and
   tracking of objects”, Proceedings from XV Brazilian Symposium on
   Computer Graphics and Image Processing, 2002. `(pdf)`_
-  Principles of Digital Image Processing: Core Algorithms
   (Undergraduate Topics in Computer Science): `Section 2.4.3 -
   Statistical shape properties`_
-  Chrystal P (1885), “On the problem to construct the minimum circle
   enclosing n given points in a plane”, *Proceedings of the Edinburgh
   Mathematical Society*, vol 3, p. 30
-  Hu MK (1962), “Visual pattern recognition by moment invariants”, *IRE
   transactions on information theory*, 8(2), pp.179-187 `(link)`_

.. _(pdf): http://sibgrapi.sid.inpe.br/col/sid.inpe.br/banon/2002/10.23.11.34/doc/35.pdf
.. _Section 2.4.3 - Statistical shape properties: http://www.scribd.com/doc/58004056/Principles-of-Digital-Image-Processing#page=49
.. _(link): https://ieeexplore.ieee.org/abstract/document/1057692
"""

"""The category of the per-object measurements made by this module"""
AREA_SHAPE = "AreaShape"

"""Calculate Zernike features for N,M where N=0 through ZERNIKE_N"""
ZERNIKE_N = 9

F_AREA = "Area"
F_PERIMETER = "Perimeter"
F_PERIMETER_CROFTON = "PerimeterCrofton"
F_VOLUME = "Volume"
F_CONVEX_VOLUME = "ConvexArea"
F_FILLED_VOLUME = "FilledArea"
F_SURFACE_AREA = "SurfaceArea"
F_ECCENTRICITY = "Eccentricity"
F_SOLIDITY = "Solidity"
F_CONVEX_AREA = "ConvexArea"
F_FILLED_AREA = "FilledArea"
F_EXTENT = "Extent"
F_CENTER_X = "Center_X"
F_CENTER_Y = "Center_Y"
F_CENTER_Z = "Center_Z"
F_BBOX_AREA = "BoundingBoxArea"
F_BBOX_VOLUME = "BoundingBoxVolume"
F_MIN_X = "BoundingBoxMinimum_X"
F_MAX_X = "BoundingBoxMaximum_X"
F_MIN_Y = "BoundingBoxMinimum_Y"
F_MAX_Y = "BoundingBoxMaximum_Y"
F_MIN_Z = "BoundingBoxMinimum_Z"
F_MAX_Z = "BoundingBoxMaximum_Z"
F_EULER_NUMBER = "EulerNumber"
F_FORM_FACTOR = "FormFactor"
F_MAJOR_AXIS_LENGTH = "MajorAxisLength"
F_MINOR_AXIS_LENGTH = "MinorAxisLength"
F_ORIENTATION = "Orientation"
F_COMPACTNESS = "Compactness"
F_INERTIA = "InertiaTensor"
F_MAXIMUM_RADIUS = "MaximumRadius"
F_MEDIAN_RADIUS = "MedianRadius"
F_MEAN_RADIUS = "MeanRadius"
F_MIN_FERET_DIAMETER = "MinFeretDiameter"
F_MAX_FERET_DIAMETER = "MaxFeretDiameter"

F_CENTRAL_MOMENT_0_0 = "CentralMoment_0_0"
F_CENTRAL_MOMENT_0_1 = "CentralMoment_0_1"
F_CENTRAL_MOMENT_0_2 = "CentralMoment_0_2"
F_CENTRAL_MOMENT_0_3 = "CentralMoment_0_3"
F_CENTRAL_MOMENT_1_0 = "CentralMoment_1_0"
F_CENTRAL_MOMENT_1_1 = "CentralMoment_1_1"
F_CENTRAL_MOMENT_1_2 = "CentralMoment_1_2"
F_CENTRAL_MOMENT_1_3 = "CentralMoment_1_3"
F_CENTRAL_MOMENT_2_0 = "CentralMoment_2_0"
F_CENTRAL_MOMENT_2_1 = "CentralMoment_2_1"
F_CENTRAL_MOMENT_2_2 = "CentralMoment_2_2"
F_CENTRAL_MOMENT_2_3 = "CentralMoment_2_3"
F_EQUIVALENT_DIAMETER = "EquivalentDiameter"
F_HU_MOMENT_0 = "HuMoment_0"
F_HU_MOMENT_1 = "HuMoment_1"
F_HU_MOMENT_2 = "HuMoment_2"
F_HU_MOMENT_3 = "HuMoment_3"
F_HU_MOMENT_4 = "HuMoment_4"
F_HU_MOMENT_5 = "HuMoment_5"
F_HU_MOMENT_6 = "HuMoment_6"
F_INERTIA_TENSOR_0_0 = "InertiaTensor_0_0"
F_INERTIA_TENSOR_0_1 = "InertiaTensor_0_1"
F_INERTIA_TENSOR_0_2 = "InertiaTensor_0_2"
F_INERTIA_TENSOR_1_0 = "InertiaTensor_1_0"
F_INERTIA_TENSOR_1_1 = "InertiaTensor_1_1"
F_INERTIA_TENSOR_1_2 = "InertiaTensor_1_2"
F_INERTIA_TENSOR_2_0 = "InertiaTensor_2_0"
F_INERTIA_TENSOR_2_1 = "InertiaTensor_2_1"
F_INERTIA_TENSOR_2_2 = "InertiaTensor_2_2"
F_INERTIA_TENSOR_EIGENVALUES_0 = "InertiaTensorEigenvalues_0"
F_INERTIA_TENSOR_EIGENVALUES_1 = "InertiaTensorEigenvalues_1"
F_INERTIA_TENSOR_EIGENVALUES_2 = "InertiaTensorEigenvalues_2"
F_NORMALIZED_MOMENT_0_0 = "NormalizedMoment_0_0"
F_NORMALIZED_MOMENT_0_1 = "NormalizedMoment_0_1"
F_NORMALIZED_MOMENT_0_2 = "NormalizedMoment_0_2"
F_NORMALIZED_MOMENT_0_3 = "NormalizedMoment_0_3"
F_NORMALIZED_MOMENT_1_0 = "NormalizedMoment_1_0"
F_NORMALIZED_MOMENT_1_1 = "NormalizedMoment_1_1"
F_NORMALIZED_MOMENT_1_2 = "NormalizedMoment_1_2"
F_NORMALIZED_MOMENT_1_3 = "NormalizedMoment_1_3"
F_NORMALIZED_MOMENT_2_0 = "NormalizedMoment_2_0"
F_NORMALIZED_MOMENT_2_1 = "NormalizedMoment_2_1"
F_NORMALIZED_MOMENT_2_2 = "NormalizedMoment_2_2"
F_NORMALIZED_MOMENT_2_3 = "NormalizedMoment_2_3"
F_NORMALIZED_MOMENT_3_0 = "NormalizedMoment_3_0"
F_NORMALIZED_MOMENT_3_1 = "NormalizedMoment_3_1"
F_NORMALIZED_MOMENT_3_2 = "NormalizedMoment_3_2"
F_NORMALIZED_MOMENT_3_3 = "NormalizedMoment_3_3"
F_SPATIAL_MOMENT_0_0 = "SpatialMoment_0_0"
F_SPATIAL_MOMENT_0_1 = "SpatialMoment_0_1"
F_SPATIAL_MOMENT_0_2 = "SpatialMoment_0_2"
F_SPATIAL_MOMENT_0_3 = "SpatialMoment_0_3"
F_SPATIAL_MOMENT_1_0 = "SpatialMoment_1_0"
F_SPATIAL_MOMENT_1_1 = "SpatialMoment_1_1"
F_SPATIAL_MOMENT_1_2 = "SpatialMoment_1_2"
F_SPATIAL_MOMENT_1_3 = "SpatialMoment_1_3"
F_SPATIAL_MOMENT_2_0 = "SpatialMoment_2_0"
F_SPATIAL_MOMENT_2_1 = "SpatialMoment_2_1"
F_SPATIAL_MOMENT_2_2 = "SpatialMoment_2_2"
F_SPATIAL_MOMENT_2_3 = "SpatialMoment_2_3"

F_SPATIAL_MOMENT_0_0_0 = "SpatialMoment_0_0_0"
F_SPATIAL_MOMENT_0_0_1 = "SpatialMoment_0_0_1"
F_SPATIAL_MOMENT_0_0_2 = "SpatialMoment_0_0_2"
F_SPATIAL_MOMENT_0_0_3 = "SpatialMoment_0_0_3"
F_SPATIAL_MOMENT_0_1_0 = "SpatialMoment_0_1_0"
F_SPATIAL_MOMENT_0_1_1 = "SpatialMoment_0_1_1"
F_SPATIAL_MOMENT_0_1_2 = "SpatialMoment_0_1_2"
F_SPATIAL_MOMENT_0_1_3 = "SpatialMoment_0_1_3"
F_SPATIAL_MOMENT_0_2_0 = "SpatialMoment_0_2_0"
F_SPATIAL_MOMENT_0_2_1 = "SpatialMoment_0_2_1"
F_SPATIAL_MOMENT_0_2_2 = "SpatialMoment_0_2_2"
F_SPATIAL_MOMENT_0_2_3 = "SpatialMoment_0_2_3"
F_SPATIAL_MOMENT_0_3_0 = "SpatialMoment_0_3_0"
F_SPATIAL_MOMENT_0_3_1 = "SpatialMoment_0_3_1"
F_SPATIAL_MOMENT_0_3_2 = "SpatialMoment_0_3_2"
F_SPATIAL_MOMENT_0_3_3 = "SpatialMoment_0_3_3"
F_SPATIAL_MOMENT_1_0_0 = "SpatialMoment_1_0_0"
F_SPATIAL_MOMENT_1_0_1 = "SpatialMoment_1_0_1"
F_SPATIAL_MOMENT_1_0_2 = "SpatialMoment_1_0_2"
F_SPATIAL_MOMENT_1_0_3 = "SpatialMoment_1_0_3"
F_SPATIAL_MOMENT_1_1_0 = "SpatialMoment_1_1_0"
F_SPATIAL_MOMENT_1_1_1 = "SpatialMoment_1_1_1"
F_SPATIAL_MOMENT_1_1_2 = "SpatialMoment_1_1_2"
F_SPATIAL_MOMENT_1_1_3 = "SpatialMoment_1_1_3"
F_SPATIAL_MOMENT_1_2_0 = "SpatialMoment_1_2_0"
F_SPATIAL_MOMENT_1_2_1 = "SpatialMoment_1_2_1"
F_SPATIAL_MOMENT_1_2_2 = "SpatialMoment_1_2_2"
F_SPATIAL_MOMENT_1_2_3 = "SpatialMoment_1_2_3"
F_SPATIAL_MOMENT_1_3_0 = "SpatialMoment_1_3_0"
F_SPATIAL_MOMENT_1_3_1 = "SpatialMoment_1_3_1"
F_SPATIAL_MOMENT_1_3_2 = "SpatialMoment_1_3_2"
F_SPATIAL_MOMENT_1_3_3 = "SpatialMoment_1_3_3"
F_SPATIAL_MOMENT_2_0_0 = "SpatialMoment_2_0_0"
F_SPATIAL_MOMENT_2_0_1 = "SpatialMoment_2_0_1"
F_SPATIAL_MOMENT_2_0_2 = "SpatialMoment_2_0_2"
F_SPATIAL_MOMENT_2_0_3 = "SpatialMoment_2_0_3"
F_SPATIAL_MOMENT_2_1_0 = "SpatialMoment_2_1_0"
F_SPATIAL_MOMENT_2_1_1 = "SpatialMoment_2_1_1"
F_SPATIAL_MOMENT_2_1_2 = "SpatialMoment_2_1_2"
F_SPATIAL_MOMENT_2_1_3 = "SpatialMoment_2_1_3"
F_SPATIAL_MOMENT_2_2_0 = "SpatialMoment_2_2_0"
F_SPATIAL_MOMENT_2_2_1 = "SpatialMoment_2_2_1"
F_SPATIAL_MOMENT_2_2_2 = "SpatialMoment_2_2_2"
F_SPATIAL_MOMENT_2_2_3 = "SpatialMoment_2_2_3"
F_SPATIAL_MOMENT_2_3_0 = "SpatialMoment_2_3_0"
F_SPATIAL_MOMENT_2_3_1 = "SpatialMoment_2_3_1"
F_SPATIAL_MOMENT_2_3_2 = "SpatialMoment_2_3_2"
F_SPATIAL_MOMENT_2_3_3 = "SpatialMoment_2_3_3"
F_SPATIAL_MOMENT_3_0_0 = "SpatialMoment_3_0_0"
F_SPATIAL_MOMENT_3_0_1 = "SpatialMoment_3_0_1"
F_SPATIAL_MOMENT_3_0_2 = "SpatialMoment_3_0_2"
F_SPATIAL_MOMENT_3_0_3 = "SpatialMoment_3_0_3"
F_SPATIAL_MOMENT_3_1_0 = "SpatialMoment_3_1_0"
F_SPATIAL_MOMENT_3_1_1 = "SpatialMoment_3_1_1"
F_SPATIAL_MOMENT_3_1_2 = "SpatialMoment_3_1_2"
F_SPATIAL_MOMENT_3_1_3 = "SpatialMoment_3_1_3"
F_SPATIAL_MOMENT_3_2_0 = "SpatialMoment_3_2_0"
F_SPATIAL_MOMENT_3_2_1 = "SpatialMoment_3_2_1"
F_SPATIAL_MOMENT_3_2_2 = "SpatialMoment_3_2_2"
F_SPATIAL_MOMENT_3_2_3 = "SpatialMoment_3_2_3"
F_SPATIAL_MOMENT_3_3_0 = "SpatialMoment_3_3_0"
F_SPATIAL_MOMENT_3_3_1 = "SpatialMoment_3_3_1"
F_SPATIAL_MOMENT_3_3_2 = "SpatialMoment_3_3_2"
F_SPATIAL_MOMENT_3_3_3 = "SpatialMoment_3_3_3"

F_NORMALIZED_MOMENT_0_0_0 = "NormalizedMoment_0_0_0"
F_NORMALIZED_MOMENT_0_0_1 = "NormalizedMoment_0_0_1"
F_NORMALIZED_MOMENT_0_0_2 = "NormalizedMoment_0_0_2"
F_NORMALIZED_MOMENT_0_0_3 = "NormalizedMoment_0_0_3"
F_NORMALIZED_MOMENT_0_1_0 = "NormalizedMoment_0_1_0"
F_NORMALIZED_MOMENT_0_1_1 = "NormalizedMoment_0_1_1"
F_NORMALIZED_MOMENT_0_1_2 = "NormalizedMoment_0_1_2"
F_NORMALIZED_MOMENT_0_1_3 = "NormalizedMoment_0_1_3"
F_NORMALIZED_MOMENT_0_2_0 = "NormalizedMoment_0_2_0"
F_NORMALIZED_MOMENT_0_2_1 = "NormalizedMoment_0_2_1"
F_NORMALIZED_MOMENT_0_2_2 = "NormalizedMoment_0_2_2"
F_NORMALIZED_MOMENT_0_2_3 = "NormalizedMoment_0_2_3"
F_NORMALIZED_MOMENT_1_0_0 = "NormalizedMoment_1_0_0"
F_NORMALIZED_MOMENT_1_0_1 = "NormalizedMoment_1_0_1"
F_NORMALIZED_MOMENT_1_0_2 = "NormalizedMoment_1_0_2"
F_NORMALIZED_MOMENT_1_0_3 = "NormalizedMoment_1_0_3"
F_NORMALIZED_MOMENT_1_1_0 = "NormalizedMoment_1_1_0"
F_NORMALIZED_MOMENT_1_1_1 = "NormalizedMoment_1_1_1"
F_NORMALIZED_MOMENT_1_1_2 = "NormalizedMoment_1_1_2"
F_NORMALIZED_MOMENT_1_1_3 = "NormalizedMoment_1_1_3"
F_NORMALIZED_MOMENT_1_2_0 = "NormalizedMoment_1_2_0"
F_NORMALIZED_MOMENT_1_2_1 = "NormalizedMoment_1_2_1"
F_NORMALIZED_MOMENT_1_2_2 = "NormalizedMoment_1_2_2"
F_NORMALIZED_MOMENT_1_2_3 = "NormalizedMoment_1_2_3"
F_NORMALIZED_MOMENT_1_3_0 = "NormalizedMoment_1_3_0"
F_NORMALIZED_MOMENT_1_3_1 = "NormalizedMoment_1_3_1"
F_NORMALIZED_MOMENT_1_3_2 = "NormalizedMoment_1_3_2"
F_NORMALIZED_MOMENT_1_3_3 = "NormalizedMoment_1_3_3"
F_NORMALIZED_MOMENT_2_0_0 = "NormalizedMoment_2_0_0"
F_NORMALIZED_MOMENT_2_0_1 = "NormalizedMoment_2_0_1"
F_NORMALIZED_MOMENT_2_0_2 = "NormalizedMoment_2_0_2"
F_NORMALIZED_MOMENT_2_0_3 = "NormalizedMoment_2_0_3"
F_NORMALIZED_MOMENT_2_1_0 = "NormalizedMoment_2_1_0"
F_NORMALIZED_MOMENT_2_1_1 = "NormalizedMoment_2_1_1"
F_NORMALIZED_MOMENT_2_1_2 = "NormalizedMoment_2_1_2"
F_NORMALIZED_MOMENT_2_1_3 = "NormalizedMoment_2_1_3"
F_NORMALIZED_MOMENT_2_2_0 = "NormalizedMoment_2_2_0"
F_NORMALIZED_MOMENT_2_2_1 = "NormalizedMoment_2_2_1"
F_NORMALIZED_MOMENT_2_2_2 = "NormalizedMoment_2_2_2"
F_NORMALIZED_MOMENT_2_2_3 = "NormalizedMoment_2_2_3"
F_NORMALIZED_MOMENT_2_3_0 = "NormalizedMoment_2_3_0"
F_NORMALIZED_MOMENT_2_3_1 = "NormalizedMoment_2_3_1"
F_NORMALIZED_MOMENT_2_3_2 = "NormalizedMoment_2_3_2"
F_NORMALIZED_MOMENT_2_3_3 = "NormalizedMoment_2_3_3"
F_NORMALIZED_MOMENT_3_0_0 = "NormalizedMoment_3_0_0"
F_NORMALIZED_MOMENT_3_0_1 = "NormalizedMoment_3_0_1"
F_NORMALIZED_MOMENT_3_0_2 = "NormalizedMoment_3_0_2"
F_NORMALIZED_MOMENT_3_0_3 = "NormalizedMoment_3_0_3"
F_NORMALIZED_MOMENT_3_1_0 = "NormalizedMoment_3_1_0"
F_NORMALIZED_MOMENT_3_1_1 = "NormalizedMoment_3_1_1"
F_NORMALIZED_MOMENT_3_1_2 = "NormalizedMoment_3_1_2"
F_NORMALIZED_MOMENT_3_1_3 = "NormalizedMoment_3_1_3"
F_NORMALIZED_MOMENT_3_2_0 = "NormalizedMoment_3_2_0"
F_NORMALIZED_MOMENT_3_2_1 = "NormalizedMoment_3_2_1"
F_NORMALIZED_MOMENT_3_2_2 = "NormalizedMoment_3_2_2"
F_NORMALIZED_MOMENT_3_2_3 = "NormalizedMoment_3_2_3"
F_NORMALIZED_MOMENT_3_3_0 = "NormalizedMoment_3_3_0"
F_NORMALIZED_MOMENT_3_3_1 = "NormalizedMoment_3_3_1"
F_NORMALIZED_MOMENT_3_3_2 = "NormalizedMoment_3_3_2"
F_NORMALIZED_MOMENT_3_3_3 = "NormalizedMoment_3_3_3"

F_CENTRAL_MOMENT_0_0_0 = "CentralMoment_0_0_0"
F_CENTRAL_MOMENT_0_0_1 = "CentralMoment_0_0_1"
F_CENTRAL_MOMENT_0_0_2 = "CentralMoment_0_0_2"
F_CENTRAL_MOMENT_0_0_3 = "CentralMoment_0_0_3"
F_CENTRAL_MOMENT_0_1_0 = "CentralMoment_0_1_0"
F_CENTRAL_MOMENT_0_1_1 = "CentralMoment_0_1_1"
F_CENTRAL_MOMENT_0_1_2 = "CentralMoment_0_1_2"
F_CENTRAL_MOMENT_0_1_3 = "CentralMoment_0_1_3"
F_CENTRAL_MOMENT_0_2_0 = "CentralMoment_0_2_0"
F_CENTRAL_MOMENT_0_2_1 = "CentralMoment_0_2_1"
F_CENTRAL_MOMENT_0_2_2 = "CentralMoment_0_2_2"
F_CENTRAL_MOMENT_0_2_3 = "CentralMoment_0_2_3"
F_CENTRAL_MOMENT_1_0_0 = "CentralMoment_1_0_0"
F_CENTRAL_MOMENT_1_0_1 = "CentralMoment_1_0_1"
F_CENTRAL_MOMENT_1_0_2 = "CentralMoment_1_0_2"
F_CENTRAL_MOMENT_1_0_3 = "CentralMoment_1_0_3"
F_CENTRAL_MOMENT_1_1_0 = "CentralMoment_1_1_0"
F_CENTRAL_MOMENT_1_1_1 = "CentralMoment_1_1_1"
F_CENTRAL_MOMENT_1_1_2 = "CentralMoment_1_1_2"
F_CENTRAL_MOMENT_1_1_3 = "CentralMoment_1_1_3"
F_CENTRAL_MOMENT_1_2_0 = "CentralMoment_1_2_0"
F_CENTRAL_MOMENT_1_2_1 = "CentralMoment_1_2_1"
F_CENTRAL_MOMENT_1_2_2 = "CentralMoment_1_2_2"
F_CENTRAL_MOMENT_1_2_3 = "CentralMoment_1_2_3"
F_CENTRAL_MOMENT_1_3_0 = "CentralMoment_1_3_0"
F_CENTRAL_MOMENT_1_3_1 = "CentralMoment_1_3_1"
F_CENTRAL_MOMENT_1_3_2 = "CentralMoment_1_3_2"
F_CENTRAL_MOMENT_1_3_3 = "CentralMoment_1_3_3"
F_CENTRAL_MOMENT_2_0_0 = "CentralMoment_2_0_0"
F_CENTRAL_MOMENT_2_0_1 = "CentralMoment_2_0_1"
F_CENTRAL_MOMENT_2_0_2 = "CentralMoment_2_0_2"
F_CENTRAL_MOMENT_2_0_3 = "CentralMoment_2_0_3"
F_CENTRAL_MOMENT_2_1_0 = "CentralMoment_2_1_0"
F_CENTRAL_MOMENT_2_1_1 = "CentralMoment_2_1_1"
F_CENTRAL_MOMENT_2_1_2 = "CentralMoment_2_1_2"
F_CENTRAL_MOMENT_2_1_3 = "CentralMoment_2_1_3"
F_CENTRAL_MOMENT_2_2_0 = "CentralMoment_2_2_0"
F_CENTRAL_MOMENT_2_2_1 = "CentralMoment_2_2_1"
F_CENTRAL_MOMENT_2_2_2 = "CentralMoment_2_2_2"
F_CENTRAL_MOMENT_2_2_3 = "CentralMoment_2_2_3"
F_CENTRAL_MOMENT_2_3_0 = "CentralMoment_2_3_0"
F_CENTRAL_MOMENT_2_3_1 = "CentralMoment_2_3_1"
F_CENTRAL_MOMENT_2_3_2 = "CentralMoment_2_3_2"
F_CENTRAL_MOMENT_2_3_3 = "CentralMoment_2_3_3"
F_CENTRAL_MOMENT_3_0_0 = "CentralMoment_3_0_0"
F_CENTRAL_MOMENT_3_0_1 = "CentralMoment_3_0_1"
F_CENTRAL_MOMENT_3_0_2 = "CentralMoment_3_0_2"
F_CENTRAL_MOMENT_3_0_3 = "CentralMoment_3_0_3"
F_CENTRAL_MOMENT_3_1_0 = "CentralMoment_3_1_0"
F_CENTRAL_MOMENT_3_1_1 = "CentralMoment_3_1_1"
F_CENTRAL_MOMENT_3_1_2 = "CentralMoment_3_1_2"
F_CENTRAL_MOMENT_3_1_3 = "CentralMoment_3_1_3"
F_CENTRAL_MOMENT_3_2_0 = "CentralMoment_3_2_0"
F_CENTRAL_MOMENT_3_2_1 = "CentralMoment_3_2_1"
F_CENTRAL_MOMENT_3_2_2 = "CentralMoment_3_2_2"
F_CENTRAL_MOMENT_3_2_3 = "CentralMoment_3_2_3"
F_CENTRAL_MOMENT_3_3_0 = "CentralMoment_3_3_0"
F_CENTRAL_MOMENT_3_3_1 = "CentralMoment_3_3_1"
F_CENTRAL_MOMENT_3_3_2 = "CentralMoment_3_3_2"
F_CENTRAL_MOMENT_3_3_3 = "CentralMoment_3_3_3"

"""The non-Zernike features"""
F_STD_2D = [
    F_AREA,
    F_PERIMETER,
    F_MAXIMUM_RADIUS,
    F_MEAN_RADIUS,
    F_MEDIAN_RADIUS,
    F_MIN_FERET_DIAMETER,
    F_MAX_FERET_DIAMETER,
    F_ORIENTATION,
    F_ECCENTRICITY,
    F_FORM_FACTOR,
    F_SOLIDITY,
    F_CONVEX_AREA,
    F_COMPACTNESS,
    F_BBOX_AREA,
]
F_STD_3D = [
    F_VOLUME,
    F_SURFACE_AREA,
    F_CENTER_Z,
    F_BBOX_VOLUME,
    F_MIN_Z,
    F_MAX_Z,
]
F_ADV_2D = [
    F_SPATIAL_MOMENT_0_0,
    F_SPATIAL_MOMENT_0_1,
    F_SPATIAL_MOMENT_0_2,
    F_SPATIAL_MOMENT_0_3,
    F_SPATIAL_MOMENT_1_0,
    F_SPATIAL_MOMENT_1_1,
    F_SPATIAL_MOMENT_1_2,
    F_SPATIAL_MOMENT_1_3,
    F_SPATIAL_MOMENT_2_0,
    F_SPATIAL_MOMENT_2_1,
    F_SPATIAL_MOMENT_2_2,
    F_SPATIAL_MOMENT_2_3,
    F_CENTRAL_MOMENT_0_0,
    F_CENTRAL_MOMENT_0_1,
    F_CENTRAL_MOMENT_0_2,
    F_CENTRAL_MOMENT_0_3,
    F_CENTRAL_MOMENT_1_0,
    F_CENTRAL_MOMENT_1_1,
    F_CENTRAL_MOMENT_1_2,
    F_CENTRAL_MOMENT_1_3,
    F_CENTRAL_MOMENT_2_0,
    F_CENTRAL_MOMENT_2_1,
    F_CENTRAL_MOMENT_2_2,
    F_CENTRAL_MOMENT_2_3,
    F_NORMALIZED_MOMENT_0_0,
    F_NORMALIZED_MOMENT_0_1,
    F_NORMALIZED_MOMENT_0_2,
    F_NORMALIZED_MOMENT_0_3,
    F_NORMALIZED_MOMENT_1_0,
    F_NORMALIZED_MOMENT_1_1,
    F_NORMALIZED_MOMENT_1_2,
    F_NORMALIZED_MOMENT_1_3,
    F_NORMALIZED_MOMENT_2_0,
    F_NORMALIZED_MOMENT_2_1,
    F_NORMALIZED_MOMENT_2_2,
    F_NORMALIZED_MOMENT_2_3,
    F_NORMALIZED_MOMENT_3_0,
    F_NORMALIZED_MOMENT_3_1,
    F_NORMALIZED_MOMENT_3_2,
    F_NORMALIZED_MOMENT_3_3,
    F_HU_MOMENT_0,
    F_HU_MOMENT_1,
    F_HU_MOMENT_2,
    F_HU_MOMENT_3,
    F_HU_MOMENT_4,
    F_HU_MOMENT_5,
    F_HU_MOMENT_6,
    F_INERTIA_TENSOR_0_0,
    F_INERTIA_TENSOR_0_1,
    F_INERTIA_TENSOR_1_0,
    F_INERTIA_TENSOR_1_1,
    F_INERTIA_TENSOR_EIGENVALUES_0,
    F_INERTIA_TENSOR_EIGENVALUES_1,
]
F_ADV_3D = [F_SOLIDITY]
F_STANDARD = [
    F_EXTENT,
    F_EULER_NUMBER,
    F_EQUIVALENT_DIAMETER,
    F_MAJOR_AXIS_LENGTH,
    F_MINOR_AXIS_LENGTH,
    F_CENTER_X,
    F_CENTER_Y,
    F_MIN_X,
    F_MIN_Y,
    F_MAX_X,
    F_MAX_Y,
]

"""
calculate_advanced : bool, optional
    If True calculate additional statistics for object moments 
    and intertia tensors in **2D mode**. These features should not require much 
    additional time to calculate, but do add many additional columns to the 
    resulting output files.

    For 3D images this setting enables the Solidity measurement, which can be 
    time-consuming to calculate.
    
calculate_zernikes : bool, optional
    If True calculate the Zernike shape features. Because the first 10 
    Zernike polynomials (from order 0 to order 9) are calculated, this operation 
    can be time consuming if the image contains a lot of objects. Set as False
    if you are measuring 3D objects with this module.
    
new_features : bool, optional
   If true calculate additional features not present in CellProfiler4 (e.g., perimeter_crofton).
"""


def get_sizeshape(
    masks: numpy.ndarray,
    pixels: numpy.ndarray,
    calculate_advanced: bool = True,
    new_features: bool = True,
    spacing: Optional[tuple] = None,
):
    """Compute the measurements for multiple object masks."""
    # Properties available for both 2d and 3d
    desired_properties = [
        "image",
        "area",
        "area_bbox",
        "area_convex",
        "equivalent_diameter_area",
        "bbox",
        "centroid",
        "euler_number",
        "extent",
        "axis_major_length",
        "axis_minor_length",
    ]

    # Features not in CellProfiler 4
    if new_features:
        desired_properties += ["area_filled"]

    # 2d specific properties
    if masks.ndim == 2:
        desired_properties += [
            "eccentricity",
            "orientation",
            "perimeter",
            "solidity",
        ]
        if new_features:
            desired_properties += ["perimeter_crofton"]

    if calculate_advanced:
        if masks.ndim == 2:
            desired_properties += [
                "inertia_tensor",
                "inertia_tensor_eigvals",
                "moments",
                "moments_hu",
                "moments_central",
                "moments_normalized",
            ]

        else:
            desired_properties += ["solidity"]

            # These advanced props were not in CP4 for 3D images
            if new_features:
                desired_properties += [
                    "inertia_tensor",
                    "inertia_tensor_eigvals",
                    "moments",
                    "moments_central",
                    "moments_normalized",
                ]

    labels = masks
    nobjects = (numpy.unique(masks) > 0).sum()
    results = {}
    if labels.ndim == 2:
        props = skimage.measure.regionprops_table(
            labels, pixels, properties=desired_properties
        )

        formfactor = 4.0 * numpy.pi * props["area"] / props["perimeter"] ** 2
        denom = [max(x, 1) for x in 4.0 * numpy.pi * props["area"]]
        compactness = props["perimeter"] ** 2 / denom

        max_radius = numpy.zeros(nobjects)
        median_radius = numpy.zeros(nobjects)
        mean_radius = numpy.zeros(nobjects)
        for index, mini_image in enumerate(props["image"]):
            # Pad image to assist distance tranform
            mini_image = numpy.pad(mini_image, 1)
            distances = scipy.ndimage.distance_transform_edt(mini_image)

            max_radius[index] = _ensure_np_scalar(
                scipy.ndimage.maximum(distances, mini_image)
            )
            mean_radius[index] = _ensure_np_scalar(
                scipy.ndimage.mean(distances, mini_image)
            )
            median_radius[index] = _ensure_np_scalar(
                centrosome.cpmorphology.median_of_labels(
                    distances, mini_image.astype("int"), [1]
                )
            )

        results |= {
            F_AREA: props["area"],
            F_BBOX_AREA: props["area_bbox"],
            F_CONVEX_AREA: props["area_convex"],
            F_EQUIVALENT_DIAMETER: props["equivalent_diameter_area"],
            F_PERIMETER: props["perimeter"],
            F_MAJOR_AXIS_LENGTH: props["axis_major_length"],
            F_MINOR_AXIS_LENGTH: props["axis_minor_length"],
            F_ECCENTRICITY: props["eccentricity"],
            F_ORIENTATION: props["orientation"] * (180 / numpy.pi),
            F_CENTER_X: props["centroid-1"],
            F_CENTER_Y: props["centroid-0"],
            F_MIN_X: props["bbox-1"],
            F_MAX_X: props["bbox-3"],
            F_MIN_Y: props["bbox-0"],
            F_MAX_Y: props["bbox-2"],
            F_FORM_FACTOR: formfactor,
            F_EXTENT: props["extent"],
            F_SOLIDITY: props["solidity"],
            F_COMPACTNESS: compactness,
            F_EULER_NUMBER: props["euler_number"],
            F_MAXIMUM_RADIUS: max_radius,
            F_MEAN_RADIUS: mean_radius,
            F_MEDIAN_RADIUS: median_radius,
        }

        if new_features:
            results |= {
                F_CONVEX_AREA: props["area_convex"],
                F_FILLED_AREA: props["area_filled"],
            }

        if calculate_advanced:
            results |= {
                F_SPATIAL_MOMENT_0_0: props["moments-0-0"],
                F_SPATIAL_MOMENT_0_1: props["moments-0-1"],
                F_SPATIAL_MOMENT_0_2: props["moments-0-2"],
                F_SPATIAL_MOMENT_0_3: props["moments-0-3"],
                F_SPATIAL_MOMENT_1_0: props["moments-1-0"],
                F_SPATIAL_MOMENT_1_1: props["moments-1-1"],
                F_SPATIAL_MOMENT_1_2: props["moments-1-2"],
                F_SPATIAL_MOMENT_1_3: props["moments-1-3"],
                F_SPATIAL_MOMENT_2_0: props["moments-2-0"],
                F_SPATIAL_MOMENT_2_1: props["moments-2-1"],
                F_SPATIAL_MOMENT_2_2: props["moments-2-2"],
                F_SPATIAL_MOMENT_2_3: props["moments-2-3"],
                F_CENTRAL_MOMENT_0_0: props["moments_central-0-0"],
                F_CENTRAL_MOMENT_0_1: props["moments_central-0-1"],
                F_CENTRAL_MOMENT_0_2: props["moments_central-0-2"],
                F_CENTRAL_MOMENT_0_3: props["moments_central-0-3"],
                F_CENTRAL_MOMENT_1_0: props["moments_central-1-0"],
                F_CENTRAL_MOMENT_1_1: props["moments_central-1-1"],
                F_CENTRAL_MOMENT_1_2: props["moments_central-1-2"],
                F_CENTRAL_MOMENT_1_3: props["moments_central-1-3"],
                F_CENTRAL_MOMENT_2_0: props["moments_central-2-0"],
                F_CENTRAL_MOMENT_2_1: props["moments_central-2-1"],
                F_CENTRAL_MOMENT_2_2: props["moments_central-2-2"],
                F_CENTRAL_MOMENT_2_3: props["moments_central-2-3"],
                F_NORMALIZED_MOMENT_0_0: props["moments_normalized-0-0"],
                F_NORMALIZED_MOMENT_0_1: props["moments_normalized-0-1"],
                F_NORMALIZED_MOMENT_0_2: props["moments_normalized-0-2"],
                F_NORMALIZED_MOMENT_0_3: props["moments_normalized-0-3"],
                F_NORMALIZED_MOMENT_1_0: props["moments_normalized-1-0"],
                F_NORMALIZED_MOMENT_1_1: props["moments_normalized-1-1"],
                F_NORMALIZED_MOMENT_1_2: props["moments_normalized-1-2"],
                F_NORMALIZED_MOMENT_1_3: props["moments_normalized-1-3"],
                F_NORMALIZED_MOMENT_2_0: props["moments_normalized-2-0"],
                F_NORMALIZED_MOMENT_2_1: props["moments_normalized-2-1"],
                F_NORMALIZED_MOMENT_2_2: props["moments_normalized-2-2"],
                F_NORMALIZED_MOMENT_2_3: props["moments_normalized-2-3"],
                F_NORMALIZED_MOMENT_3_0: props["moments_normalized-3-0"],
                F_NORMALIZED_MOMENT_3_1: props["moments_normalized-3-1"],
                F_NORMALIZED_MOMENT_3_2: props["moments_normalized-3-2"],
                F_NORMALIZED_MOMENT_3_3: props["moments_normalized-3-3"],
                F_HU_MOMENT_0: props["moments_hu-0"],
                F_HU_MOMENT_1: props["moments_hu-1"],
                F_HU_MOMENT_2: props["moments_hu-2"],
                F_HU_MOMENT_3: props["moments_hu-3"],
                F_HU_MOMENT_4: props["moments_hu-4"],
                F_HU_MOMENT_5: props["moments_hu-5"],
                F_HU_MOMENT_6: props["moments_hu-6"],
                F_INERTIA_TENSOR_0_0: props["inertia_tensor-0-0"],
                F_INERTIA_TENSOR_0_1: props["inertia_tensor-0-1"],
                F_INERTIA_TENSOR_1_0: props["inertia_tensor-1-0"],
                F_INERTIA_TENSOR_1_1: props["inertia_tensor-1-1"],
                F_INERTIA_TENSOR_EIGENVALUES_0: props["inertia_tensor_eigvals-0"],
                F_INERTIA_TENSOR_EIGENVALUES_1: props["inertia_tensor_eigvals-1"],
            }

        if new_features:
            results |= {
                F_PERIMETER_CROFTON: props["perimeter_crofton"],
                F_FILLED_AREA: props["area_filled"],
            }

    else:
        props = skimage.measure.regionprops_table(
            labels, pixels, properties=desired_properties
        )

        # SurfaceArea
        distinct_labels = numpy.unique(labels[labels > 0])
        surface_areas = numpy.zeros(len(distinct_labels))
        for idx, label in enumerate(distinct_labels):
            # this seems less elegant than you might wish, given that regionprops returns a slice,
            # but we need to expand the slice out by one voxel in each direction, or surface area freaks out
            slices = tuple(
                slice(
                    max(props[f"bbox-{i}"][idx] - 1, 0),
                    min(props[f"bbox-{i + 3}"][idx] + 1, labels.shape[i]),
                )
                for i in range(3)
            )
            volume = labels[slices] == label
            spacing = spacing if spacing is not None else (1.0,) * labels.ndim
            verts, faces, _, _ = skimage.measure.marching_cubes(
                volume,
                method="lewiner",
                spacing=spacing,
                level=0,
            )
            surface_areas[idx] = skimage.measure.mesh_surface_area(verts, faces)

        results = results | {
            F_VOLUME: props["area"],
            F_SURFACE_AREA: surface_areas,
            F_MAJOR_AXIS_LENGTH: props["axis_major_length"],
            F_MINOR_AXIS_LENGTH: props["axis_minor_length"],
            F_CENTER_X: props["centroid-2"],
            F_CENTER_Y: props["centroid-1"],
            F_CENTER_Z: props["centroid-0"],
            F_BBOX_VOLUME: props["area_bbox"],
            F_MIN_X: props["bbox-2"],
            F_MAX_X: props["bbox-5"],
            F_MIN_Y: props["bbox-1"],
            F_MAX_Y: props["bbox-4"],
            F_MIN_Z: props["bbox-0"],
            F_MAX_Z: props["bbox-3"],
            F_EXTENT: props["extent"],
            F_EULER_NUMBER: props["euler_number"],
            F_EQUIVALENT_DIAMETER: props["equivalent_diameter_area"],
        }
        if calculate_advanced:
            results |= {
                F_SOLIDITY: props["solidity"],
            }
            if new_features:
                results |= {
                    F_SPATIAL_MOMENT_0_0_0: props["moments-0-0-0"],
                    F_SPATIAL_MOMENT_0_0_1: props["moments-0-0-1"],
                    F_SPATIAL_MOMENT_0_0_2: props["moments-0-0-2"],
                    F_SPATIAL_MOMENT_0_0_3: props["moments-0-0-3"],
                    F_SPATIAL_MOMENT_0_1_0: props["moments-0-1-0"],
                    F_SPATIAL_MOMENT_0_1_1: props["moments-0-1-1"],
                    F_SPATIAL_MOMENT_0_1_2: props["moments-0-1-2"],
                    F_SPATIAL_MOMENT_0_1_3: props["moments-0-1-3"],
                    F_SPATIAL_MOMENT_0_2_0: props["moments-0-2-0"],
                    F_SPATIAL_MOMENT_0_2_1: props["moments-0-2-1"],
                    F_SPATIAL_MOMENT_0_2_2: props["moments-0-2-2"],
                    F_SPATIAL_MOMENT_0_2_3: props["moments-0-2-3"],
                    F_SPATIAL_MOMENT_1_0_0: props["moments-1-0-0"],
                    F_SPATIAL_MOMENT_1_0_1: props["moments-1-0-1"],
                    F_SPATIAL_MOMENT_1_0_2: props["moments-1-0-2"],
                    F_SPATIAL_MOMENT_1_0_3: props["moments-1-0-3"],
                    F_SPATIAL_MOMENT_1_1_0: props["moments-1-1-0"],
                    F_SPATIAL_MOMENT_1_1_1: props["moments-1-1-1"],
                    F_SPATIAL_MOMENT_1_1_2: props["moments-1-1-2"],
                    F_SPATIAL_MOMENT_1_1_3: props["moments-1-1-3"],
                    F_SPATIAL_MOMENT_1_2_0: props["moments-1-2-0"],
                    F_SPATIAL_MOMENT_1_2_1: props["moments-1-2-1"],
                    F_SPATIAL_MOMENT_1_2_2: props["moments-1-2-2"],
                    F_SPATIAL_MOMENT_1_2_3: props["moments-1-2-3"],
                    F_SPATIAL_MOMENT_1_3_0: props["moments-1-3-0"],
                    F_SPATIAL_MOMENT_1_3_1: props["moments-1-3-1"],
                    F_SPATIAL_MOMENT_1_3_2: props["moments-1-3-2"],
                    F_SPATIAL_MOMENT_1_3_3: props["moments-1-3-3"],
                    F_SPATIAL_MOMENT_2_0_0: props["moments-2-0-0"],
                    F_SPATIAL_MOMENT_2_0_1: props["moments-2-0-1"],
                    F_SPATIAL_MOMENT_2_0_2: props["moments-2-0-2"],
                    F_SPATIAL_MOMENT_2_0_3: props["moments-2-0-3"],
                    F_SPATIAL_MOMENT_2_1_0: props["moments-2-1-0"],
                    F_SPATIAL_MOMENT_2_1_1: props["moments-2-1-1"],
                    F_SPATIAL_MOMENT_2_1_2: props["moments-2-1-2"],
                    F_SPATIAL_MOMENT_2_1_3: props["moments-2-1-3"],
                    F_SPATIAL_MOMENT_2_2_0: props["moments-2-2-0"],
                    F_SPATIAL_MOMENT_2_2_1: props["moments-2-2-1"],
                    F_SPATIAL_MOMENT_2_2_2: props["moments-2-2-2"],
                    F_SPATIAL_MOMENT_2_2_3: props["moments-2-2-3"],
                    F_SPATIAL_MOMENT_2_3_0: props["moments-2-3-0"],
                    F_SPATIAL_MOMENT_2_3_1: props["moments-2-3-1"],
                    F_SPATIAL_MOMENT_2_3_2: props["moments-2-3-2"],
                    F_SPATIAL_MOMENT_2_3_3: props["moments-2-3-3"],
                    F_SPATIAL_MOMENT_3_0_0: props["moments-3-0-0"],
                    F_SPATIAL_MOMENT_3_0_1: props["moments-3-0-1"],
                    F_SPATIAL_MOMENT_3_0_2: props["moments-3-0-2"],
                    F_SPATIAL_MOMENT_3_0_3: props["moments-3-0-3"],
                    F_SPATIAL_MOMENT_3_1_0: props["moments-3-1-0"],
                    F_SPATIAL_MOMENT_3_1_1: props["moments-3-1-1"],
                    F_SPATIAL_MOMENT_3_1_2: props["moments-3-1-2"],
                    F_SPATIAL_MOMENT_3_1_3: props["moments-3-1-3"],
                    F_SPATIAL_MOMENT_3_2_0: props["moments-3-2-0"],
                    F_SPATIAL_MOMENT_3_2_1: props["moments-3-2-1"],
                    F_SPATIAL_MOMENT_3_2_2: props["moments-3-2-2"],
                    F_SPATIAL_MOMENT_3_2_3: props["moments-3-2-3"],
                    F_SPATIAL_MOMENT_3_3_0: props["moments-3-3-0"],
                    F_SPATIAL_MOMENT_3_3_1: props["moments-3-3-1"],
                    F_SPATIAL_MOMENT_3_3_2: props["moments-3-3-2"],
                    F_SPATIAL_MOMENT_3_3_3: props["moments-3-3-3"],
                    F_CENTRAL_MOMENT_0_0_0: props["moments_central-0-0-0"],
                    F_CENTRAL_MOMENT_0_0_1: props["moments_central-0-0-1"],
                    F_CENTRAL_MOMENT_0_0_2: props["moments_central-0-0-2"],
                    F_CENTRAL_MOMENT_0_0_3: props["moments_central-0-0-3"],
                    F_CENTRAL_MOMENT_0_1_0: props["moments_central-0-1-0"],
                    F_CENTRAL_MOMENT_0_1_1: props["moments_central-0-1-1"],
                    F_CENTRAL_MOMENT_0_1_2: props["moments_central-0-1-2"],
                    F_CENTRAL_MOMENT_0_1_3: props["moments_central-0-1-3"],
                    F_CENTRAL_MOMENT_0_2_0: props["moments_central-0-2-0"],
                    F_CENTRAL_MOMENT_0_2_1: props["moments_central-0-2-1"],
                    F_CENTRAL_MOMENT_0_2_2: props["moments_central-0-2-2"],
                    F_CENTRAL_MOMENT_0_2_3: props["moments_central-0-2-3"],
                    F_CENTRAL_MOMENT_1_0_0: props["moments_central-1-0-0"],
                    F_CENTRAL_MOMENT_1_0_1: props["moments_central-1-0-1"],
                    F_CENTRAL_MOMENT_1_0_2: props["moments_central-1-0-2"],
                    F_CENTRAL_MOMENT_1_0_3: props["moments_central-1-0-3"],
                    F_CENTRAL_MOMENT_1_1_0: props["moments_central-1-1-0"],
                    F_CENTRAL_MOMENT_1_1_1: props["moments_central-1-1-1"],
                    F_CENTRAL_MOMENT_1_1_2: props["moments_central-1-1-2"],
                    F_CENTRAL_MOMENT_1_1_3: props["moments_central-1-1-3"],
                    F_CENTRAL_MOMENT_1_2_0: props["moments_central-1-2-0"],
                    F_CENTRAL_MOMENT_1_2_1: props["moments_central-1-2-1"],
                    F_CENTRAL_MOMENT_1_2_2: props["moments_central-1-2-2"],
                    F_CENTRAL_MOMENT_1_2_3: props["moments_central-1-2-3"],
                    F_CENTRAL_MOMENT_1_3_0: props["moments_central-1-3-0"],
                    F_CENTRAL_MOMENT_1_3_1: props["moments_central-1-3-1"],
                    F_CENTRAL_MOMENT_1_3_2: props["moments_central-1-3-2"],
                    F_CENTRAL_MOMENT_1_3_3: props["moments_central-1-3-3"],
                    F_CENTRAL_MOMENT_2_0_0: props["moments_central-2-0-0"],
                    F_CENTRAL_MOMENT_2_0_1: props["moments_central-2-0-1"],
                    F_CENTRAL_MOMENT_2_0_2: props["moments_central-2-0-2"],
                    F_CENTRAL_MOMENT_2_0_3: props["moments_central-2-0-3"],
                    F_CENTRAL_MOMENT_2_1_0: props["moments_central-2-1-0"],
                    F_CENTRAL_MOMENT_2_1_1: props["moments_central-2-1-1"],
                    F_CENTRAL_MOMENT_2_1_2: props["moments_central-2-1-2"],
                    F_CENTRAL_MOMENT_2_1_3: props["moments_central-2-1-3"],
                    F_CENTRAL_MOMENT_2_2_0: props["moments_central-2-2-0"],
                    F_CENTRAL_MOMENT_2_2_1: props["moments_central-2-2-1"],
                    F_CENTRAL_MOMENT_2_2_2: props["moments_central-2-2-2"],
                    F_CENTRAL_MOMENT_2_2_3: props["moments_central-2-2-3"],
                    F_CENTRAL_MOMENT_2_3_0: props["moments_central-2-3-0"],
                    F_CENTRAL_MOMENT_2_3_1: props["moments_central-2-3-1"],
                    F_CENTRAL_MOMENT_2_3_2: props["moments_central-2-3-2"],
                    F_CENTRAL_MOMENT_2_3_3: props["moments_central-2-3-3"],
                    F_CENTRAL_MOMENT_3_0_0: props["moments_central-3-0-0"],
                    F_CENTRAL_MOMENT_3_0_1: props["moments_central-3-0-1"],
                    F_CENTRAL_MOMENT_3_0_2: props["moments_central-3-0-2"],
                    F_CENTRAL_MOMENT_3_0_3: props["moments_central-3-0-3"],
                    F_CENTRAL_MOMENT_3_1_0: props["moments_central-3-1-0"],
                    F_CENTRAL_MOMENT_3_1_1: props["moments_central-3-1-1"],
                    F_CENTRAL_MOMENT_3_1_2: props["moments_central-3-1-2"],
                    F_CENTRAL_MOMENT_3_1_3: props["moments_central-3-1-3"],
                    F_CENTRAL_MOMENT_3_2_0: props["moments_central-3-2-0"],
                    F_CENTRAL_MOMENT_3_2_1: props["moments_central-3-2-1"],
                    F_CENTRAL_MOMENT_3_2_2: props["moments_central-3-2-2"],
                    F_CENTRAL_MOMENT_3_2_3: props["moments_central-3-2-3"],
                    F_CENTRAL_MOMENT_3_3_0: props["moments_central-3-3-0"],
                    F_CENTRAL_MOMENT_3_3_1: props["moments_central-3-3-1"],
                    F_CENTRAL_MOMENT_3_3_2: props["moments_central-3-3-2"],
                    F_CENTRAL_MOMENT_3_3_3: props["moments_central-3-3-3"],
                    F_NORMALIZED_MOMENT_0_0_0: props["moments_normalized-0-0-0"],
                    F_NORMALIZED_MOMENT_0_0_1: props["moments_normalized-0-0-1"],
                    F_NORMALIZED_MOMENT_0_0_2: props["moments_normalized-0-0-2"],
                    F_NORMALIZED_MOMENT_0_0_3: props["moments_normalized-0-0-3"],
                    F_NORMALIZED_MOMENT_0_1_0: props["moments_normalized-0-1-0"],
                    F_NORMALIZED_MOMENT_0_1_1: props["moments_normalized-0-1-1"],
                    F_NORMALIZED_MOMENT_0_1_2: props["moments_normalized-0-1-2"],
                    F_NORMALIZED_MOMENT_0_1_3: props["moments_normalized-0-1-3"],
                    F_NORMALIZED_MOMENT_0_2_0: props["moments_normalized-0-2-0"],
                    F_NORMALIZED_MOMENT_0_2_1: props["moments_normalized-0-2-1"],
                    F_NORMALIZED_MOMENT_0_2_2: props["moments_normalized-0-2-2"],
                    F_NORMALIZED_MOMENT_0_2_3: props["moments_normalized-0-2-3"],
                    F_NORMALIZED_MOMENT_1_0_0: props["moments_normalized-1-0-0"],
                    F_NORMALIZED_MOMENT_1_0_1: props["moments_normalized-1-0-1"],
                    F_NORMALIZED_MOMENT_1_0_2: props["moments_normalized-1-0-2"],
                    F_NORMALIZED_MOMENT_1_0_3: props["moments_normalized-1-0-3"],
                    F_NORMALIZED_MOMENT_1_1_0: props["moments_normalized-1-1-0"],
                    F_NORMALIZED_MOMENT_1_1_1: props["moments_normalized-1-1-1"],
                    F_NORMALIZED_MOMENT_1_1_2: props["moments_normalized-1-1-2"],
                    F_NORMALIZED_MOMENT_1_1_3: props["moments_normalized-1-1-3"],
                    F_NORMALIZED_MOMENT_1_2_0: props["moments_normalized-1-2-0"],
                    F_NORMALIZED_MOMENT_1_2_1: props["moments_normalized-1-2-1"],
                    F_NORMALIZED_MOMENT_1_2_2: props["moments_normalized-1-2-2"],
                    F_NORMALIZED_MOMENT_1_2_3: props["moments_normalized-1-2-3"],
                    F_NORMALIZED_MOMENT_1_3_0: props["moments_normalized-1-3-0"],
                    F_NORMALIZED_MOMENT_1_3_1: props["moments_normalized-1-3-1"],
                    F_NORMALIZED_MOMENT_1_3_2: props["moments_normalized-1-3-2"],
                    F_NORMALIZED_MOMENT_1_3_3: props["moments_normalized-1-3-3"],
                    F_NORMALIZED_MOMENT_2_0_0: props["moments_normalized-2-0-0"],
                    F_NORMALIZED_MOMENT_2_0_1: props["moments_normalized-2-0-1"],
                    F_NORMALIZED_MOMENT_2_0_2: props["moments_normalized-2-0-2"],
                    F_NORMALIZED_MOMENT_2_0_3: props["moments_normalized-2-0-3"],
                    F_NORMALIZED_MOMENT_2_1_0: props["moments_normalized-2-1-0"],
                    F_NORMALIZED_MOMENT_2_1_1: props["moments_normalized-2-1-1"],
                    F_NORMALIZED_MOMENT_2_1_2: props["moments_normalized-2-1-2"],
                    F_NORMALIZED_MOMENT_2_1_3: props["moments_normalized-2-1-3"],
                    F_NORMALIZED_MOMENT_2_2_0: props["moments_normalized-2-2-0"],
                    F_NORMALIZED_MOMENT_2_2_1: props["moments_normalized-2-2-1"],
                    F_NORMALIZED_MOMENT_2_2_2: props["moments_normalized-2-2-2"],
                    F_NORMALIZED_MOMENT_2_2_3: props["moments_normalized-2-2-3"],
                    F_NORMALIZED_MOMENT_2_3_0: props["moments_normalized-2-3-0"],
                    F_NORMALIZED_MOMENT_2_3_1: props["moments_normalized-2-3-1"],
                    F_NORMALIZED_MOMENT_2_3_2: props["moments_normalized-2-3-2"],
                    F_NORMALIZED_MOMENT_2_3_3: props["moments_normalized-2-3-3"],
                    F_NORMALIZED_MOMENT_3_0_0: props["moments_normalized-3-0-0"],
                    F_NORMALIZED_MOMENT_3_0_1: props["moments_normalized-3-0-1"],
                    F_NORMALIZED_MOMENT_3_0_2: props["moments_normalized-3-0-2"],
                    F_NORMALIZED_MOMENT_3_0_3: props["moments_normalized-3-0-3"],
                    F_NORMALIZED_MOMENT_3_1_0: props["moments_normalized-3-1-0"],
                    F_NORMALIZED_MOMENT_3_1_1: props["moments_normalized-3-1-1"],
                    F_NORMALIZED_MOMENT_3_1_2: props["moments_normalized-3-1-2"],
                    F_NORMALIZED_MOMENT_3_1_3: props["moments_normalized-3-1-3"],
                    F_NORMALIZED_MOMENT_3_2_0: props["moments_normalized-3-2-0"],
                    F_NORMALIZED_MOMENT_3_2_1: props["moments_normalized-3-2-1"],
                    F_NORMALIZED_MOMENT_3_2_2: props["moments_normalized-3-2-2"],
                    F_NORMALIZED_MOMENT_3_2_3: props["moments_normalized-3-2-3"],
                    F_NORMALIZED_MOMENT_3_3_0: props["moments_normalized-3-3-0"],
                    F_NORMALIZED_MOMENT_3_3_1: props["moments_normalized-3-3-1"],
                    F_NORMALIZED_MOMENT_3_3_2: props["moments_normalized-3-3-2"],
                    F_NORMALIZED_MOMENT_3_3_3: props["moments_normalized-3-3-3"],
                    F_INERTIA_TENSOR_0_0: props["inertia_tensor-0-0"],
                    F_INERTIA_TENSOR_0_1: props["inertia_tensor-0-1"],
                    F_INERTIA_TENSOR_0_2: props["inertia_tensor-0-2"],
                    F_INERTIA_TENSOR_1_0: props["inertia_tensor-1-0"],
                    F_INERTIA_TENSOR_1_1: props["inertia_tensor-1-1"],
                    F_INERTIA_TENSOR_1_2: props["inertia_tensor-1-2"],
                    F_INERTIA_TENSOR_2_0: props["inertia_tensor-2-0"],
                    F_INERTIA_TENSOR_2_1: props["inertia_tensor-2-1"],
                    F_INERTIA_TENSOR_2_2: props["inertia_tensor-2-2"],
                    F_INERTIA_TENSOR_EIGENVALUES_0: props["inertia_tensor_eigvals-0"],
                    F_INERTIA_TENSOR_EIGENVALUES_1: props["inertia_tensor_eigvals-1"],
                    F_INERTIA_TENSOR_EIGENVALUES_2: props["inertia_tensor_eigvals-2"],
                }
        if new_features:
            results |= {
                F_FILLED_VOLUME: props["area_filled"],
                F_CONVEX_VOLUME: props["area_convex"],
            }

    return results


def get_zernike(masks: numpy.ndarray, pixels: numpy.ndarray, zernike_numbers: int = 9):
    #
    # Zernike features
    #
    unique_indices = numpy.unique(masks)
    unique_indices = unique_indices[unique_indices > 0]
    indices = list(range(1, len(unique_indices) + 1))
    labels = masks
    zernike_numbers = centrosome.zernike.get_zernike_indexes(zernike_numbers + 1)

    zf_l = centrosome.zernike.zernike(zernike_numbers, labels, indices)
    results = {}
    for (n, m), z in zip(zernike_numbers, zf_l.transpose()):
        results[f"Zernike_{n}_{m}"] = z

    return results


def get_ferret(masks: numpy.ndarray, pixels: numpy.ndarray):
    ijv = masks_to_ijv(masks)
    indices = numpy.unique(ijv[:, 2])
    indices = indices[indices > 0]
    chulls, chull_counts = centrosome.cpmorphology.convex_hull_ijv(ijv, indices)
    #
    # Feret diameter
    #
    (
        min_feret_diameter,
        max_feret_diameter,
    ) = centrosome.cpmorphology.feret_diameter(chulls, chull_counts, indices)

    features_to_record = {
        F_MIN_FERET_DIAMETER: min_feret_diameter,
        F_MAX_FERET_DIAMETER: max_feret_diameter,
    }

    return features_to_record



"""
Convenience functions for tests and demos.
"""

import numpy


def get_pixels(size: int = 789, seed: int = 42) -> numpy.ndarray:
    rng = numpy.random.default_rng(seed)
    random_pixels = rng.integers(low=0, high=10, size=(size, size))

    return random_pixels


def get_masks(size: int = 789, mask_width: int = 9) -> tuple[numpy.ndarray]:
    full = numpy.ones(
        (size, size), dtype=numpy.uint16
    )  # A mask that covers the entire image
    empty = numpy.zeros_like(full)

    # A mask in the center
    center = empty.copy()
    center[mask_width : -mask_width - 1, mask_width : -mask_width - 1] = True

    # Everything but a mask in the center
    edges = (~center.astype(bool)).astype(numpy.uint16)

    # Only corners
    corner = empty.copy()
    corner[:mask_width, :mask_width] = True
    corners = [numpy.rot90(corner, k=i, axes=(0, 1)) for i in range(4)]

    # One pixel mask
    one = empty.copy()
    one[size // 2, size // 2] = True

    # Two pixels' mask
    two = one + numpy.roll(one, shift=1)

    # Two masks
    full_two_masks = full.copy()
    full_two_masks[size // 2 : -1, :] = 2

    return {
        "full": full,
        "full_2": full_two_masks,
        "center": center,
        "edges": edges,
        **{f"corner_{i}": corners[i] for i in range(4)},
        "one": one,
        "two": two,
    }
